/*
 *	Vertical blank interrupt driver for PCI devices
 *
 *	(C) Copyright 2003 Soeren Sandmann
 *	(C) Copyright 2003 Red Hat
 *	
 *	Based heavily on svgalib interrupt.c by Matan Ziv-Av
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	0.00		SS		Original Driver
 *	0.01		AC		Moved to miscdev, rewrote interfaces
 */
 
#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/init.h>

#include <linux/vblankdev.h>

static int try_vga = 0;		/* Set to try generic VGA on unknowns */

struct video_card;

struct vblank_ops
{
	int  (*test)(struct video_card *);
	void (*enable)(struct video_card *);
	void (*disable)(struct video_card *);
};

struct video_card
{
	struct pci_dev *pci_dev;	/* PCI device */
	unsigned long iobase;		/* Base for I/O */
	unsigned long mmio;		/* Our MMIO mapping if any */
	unsigned long mmio_len;
	struct vblank_ops *vops;
	
	int running:1;
	
	unsigned int count;		/* Counter for interrupts */
	spinlock_t lock;		/* Protects count */
	
	wait_queue_head_t wait;		/* Woken each vblank */
	struct fasync_struct *async_queue;	/* Async chain */
	
	u32 saved_pmc;
};


/*
 *	Based on interrupt.c from svgalib
 *	Author: Matan Ziv-Av (matan@svgalib.org) 
 */
 
static int vga_test_vsync(struct video_card *card)
{
	return inb(card->iobase + 0x3C2) & 0x80;
}

static void vga_clear_disable_vsync(struct video_card *card)
{
	u8 pb;

	outb(0x11, card->iobase + 0x3d4);
	pb = inb(card->iobase + 0x3d5);
	outb(0x11, card->iobase + 0x3d4);
	outb(pb & 0xef, card->iobase + 0x3d5);
}

static void vga_enable_vsync(struct video_card *card)
{
	u8 pb;

	/*
	 *	Enable interrupt, clear pending 
	 */
	outb(0x11, card->iobase + 0x3d4);
	pb = inb(card->iobase + 0x3d5);
	outb(0x11, card->iobase + 0x3d4);
	outb((pb&0xcf) , card->iobase + 0x3d5);
    
	/*
	 *	Allow interrupts
	 */
	outb(0x11, card->iobase + 0x3d4);
	pb = inb(card->iobase + 0x3d5);
	outb(0x11, card->iobase + 0x3d4);
	outb(pb | 0x10 , card->iobase + 0x3d5);
}

static struct vblank_ops vblank_io_ops =
{
	vga_test_vsync,
	vga_enable_vsync,
	vga_clear_disable_vsync,
};
                                                                                
static int vga_mm_test_vsync(struct video_card *card)
{
	return readb(card->iobase + 0x3C2) & 0x80;
}

static void vga_mm_clear_disable_vsync(struct video_card *card)
{
	u8 pb;

	writeb(0x11, card->iobase + 0x3d4);
	pb = readb(card->iobase + 0x3d5);
	writeb(0x11, card->iobase + 0x3d4);
	writeb(pb & 0xef, card->iobase + 0x3d5);
}

static void vga_mm_enable_vsync(struct video_card *card)
{
	u8 pb;

	/*
	 *	Enable interrupt, clear pending 
	 */
	writeb(0x11, card->iobase + 0x3d4);
	pb = readb(card->iobase + 0x3d5);
	writeb(0x11, card->iobase + 0x3d4);
	writeb((pb&0xcf) , card->iobase + 0x3d5);
    
	/*
	 *	Allow interrupts
	 */
	writeb(0x11, card->iobase + 0x3d4);
	pb = readb(card->iobase + 0x3d5);
	writeb(0x11, card->iobase + 0x3d4);
	writeb(pb | 0x10 , card->iobase + 0x3d5);
}

static struct vblank_ops vga_mm_ops =
{
	vga_mm_test_vsync,
	vga_mm_enable_vsync,
	vga_mm_clear_disable_vsync,
};

                                                                                
static int nv3_test_vsync(struct video_card *card)
{
	return readl(card->iobase+0x400100)&0x100;
}

static void nv3_clear_and_disable_vsync(struct video_card *card)
{
	/* disable interrupt, clear pending */
	writel(0xffffffff,  card->iobase + 0x000100);
	writel(0x100, 	card->iobase + 0x400100);
	writel(0, 		card->iobase + 0x000140);
	writel(0, 		card->iobase + 0x400140);
	writel(card->saved_pmc,	card->iobase + 0x000200); 
}

static void nv3_enable_vsync(struct video_card *card)
{
	card->saved_pmc = inl(card->iobase + 0x200);
	writel(card->saved_pmc|0x1000, card->iobase+0x200);
	writel(0x1, 		card->iobase + 0x000140);
	writel(0x100, 		card->iobase + 0x400140);
	writel(0xffffffff, 		card->iobase + 0x000100);
	writel(0xffffffff,		card->iobase + 0x400100);
}

static struct vblank_ops nv3_ops = {
	nv3_test_vsync,
	nv3_enable_vsync,
	nv3_clear_and_disable_vsync,
};

static int nv4_test_vsync(struct video_card *card)
{
	 return readl(card->iobase+0x600100)&0x1;
}

static void nv4_clear_and_disable_vsync(struct video_card *card)
{
	/* disable interrupt, clear pending */
	writel(0xffffffff,  	card->iobase + 0x000100);
	writel(0x1,	 	card->iobase + 0x600100);
	writel(0, 		card->iobase + 0x000140);
	writel(0, 		card->iobase + 0x600140);
	writel(card->saved_pmc,	card->iobase + 0x000200); 
}

static void nv4_enable_vsync(struct video_card *card)
{
	card->saved_pmc = inl(card->iobase + 0x200);
	writel(card->saved_pmc|(1<<24),card->iobase+0x200);
	writel(0x1, 		card->iobase + 0x000140);
	writel(0x1, 		card->iobase + 0x600140);
	writel(0xffffffff, 	card->iobase + 0x000100);
	writel(0xffffffff, 	card->iobase + 0x600100);
}

static struct vblank_ops nv4_ops = {
	nv4_test_vsync,
	nv4_enable_vsync,
	nv4_clear_and_disable_vsync,
};


static int r128_test_vsync(struct video_card *card)
{
	return readl(card->iobase + 0x44) &1;
}

static void r128_clear_and_disable_vsync(struct video_card *card)
{
	writel(1, card->iobase + 0x44);
	writel(readl(card->iobase + 0x40) & 0xfffffffe, card->iobase + 0x40); 
}

static void r128_enable_vsync(struct video_card *card)
{
	writel(1, card->iobase + 0x44);
	writel(readl(card->iobase + 0x40) | 1, card->iobase + 0x40); 
}

static struct vblank_ops r128_ops = {
	r128_test_vsync,
	r128_enable_vsync,
	r128_clear_and_disable_vsync
};

static int rage_test_vsync(struct video_card *card)
{
	return inl(card->iobase + 0x18) &4;
}

static void rage_clear_and_disable_vsync(struct video_card *card)
{
	outl((inl(card->iobase + 0x18) & 0xfffffff8) | 4, card->iobase + 0x18); 
}

static void rage_enable_vsync(struct video_card *card)
{
	outl((inl(card->iobase + 0x18) & 0xfffffff8) | 6, card->iobase + 0x18); 
}

static struct vblank_ops rage_ops = {
	rage_test_vsync,
	rage_enable_vsync,
	rage_clear_and_disable_vsync
};

static int rendition_test_vsync(struct video_card *card)
{
	return inw(card->iobase + 0x44) & 1;
}

static void rendition_clear_and_disable_vsync(struct video_card *card)
{
	outw(1, card->iobase + 0x44);
	outw(0, card->iobase + 0x46); 
}

static void rendition_enable_vsync(struct video_card *card)
{
	outw(1, card->iobase + 0x44);
	outw(1, card->iobase + 0x46); 
}

static struct vblank_ops rendition_ops = {
	rendition_test_vsync,
	rendition_enable_vsync,
	rendition_clear_and_disable_vsync,
};

static int is_r128(struct pci_dev *pdev)
{
	switch(pdev->device)
	{
		case 0x4c45:
		case 0x4c46:
		case 0x4d46:
		case 0x4d4c:
			return 1;
		case 0x4242:
		case 0x4c57:
		case 0x4c59:
		case 0x4c5a:
			return 1;
	}
	
	switch(pdev->device >> 8)
	{
		case 0x50:
		case 0x52:
		case 0x53:
		case 0x54:
			return 1;
		case 0x51:
			return 1;
	}
	
	return 0;
		
}

static int vga_init_vsync(struct video_card *card)
{
	struct pci_dev *pdev = card->pci_dev;
	int res;
		
	switch(pdev->vendor)
	{
		case PCI_VENDOR_ID_MATROX:
			res = 0;
			if(pci_resource_len(pdev, 0) >= 1048576)
				res = 1;
			card->mmio = (unsigned long)ioremap(pci_resource_start(pdev, res), 0x2000);
			card->mmio_len = 0x2000;
			if(card->mmio == 0)
				goto fail;
			card->iobase = card->mmio + 0x1C00;
			card->vops = &vga_mm_ops;
			break;
		case PCI_VENDOR_ID_SI:
			card->iobase = pci_resource_start(pdev, 2) - 0x380;
			card->vops = &vblank_io_ops;
			break;
		case PCI_VENDOR_ID_NVIDIA_SGS:
			card->mmio = (unsigned long)ioremap(pci_resource_start(pdev, 0), 0x800000);
			card->mmio_len = 0x800000;
			if(card->mmio == 0)
				goto fail;
			card->iobase = card->mmio;
			if(pdev->device < 0x20)
				card->vops = &nv3_ops;
			else
				card->vops = &nv4_ops;
			break;
		case PCI_VENDOR_ID_NVIDIA:
			card->mmio = (unsigned long)ioremap(pci_resource_start(pdev, 0), 0x800000);
			card->mmio_len = 0x800000;
			if(card->mmio == 0)
				goto fail;
			card->iobase = card->mmio;
			card->vops= &nv4_ops;
			break;
		case PCI_VENDOR_ID_ATI:
			/* FIXME: Add Radeon */
			if(is_r128(pdev))
			{
				card->mmio = (unsigned long)ioremap(pci_resource_start(pdev, 2), 16384);
				card->mmio_len = 16384;
				if(card->mmio == 0)
					goto fail;
				card->iobase = card->mmio;
				card->vops = &r128_ops;
			}
			else
			{
				card->iobase = pci_resource_start(pdev, 1);
				card->vops = &rage_ops;
			}
			break;
		case PCI_VENDOR_ID_RENDITION:
			card->iobase = pci_resource_start(pdev, 1);
			card->vops = &rendition_ops;
			break;
		default:
			if(try_vga)
			{
				card->iobase = 0;
				card->vops = &vblank_io_ops;
				break;
			}
			printk(KERN_ERR "vblank: Unsupported video card %04X:%04X.\n",
				pdev->vendor, pdev->device);
			return -EOPNOTSUPP;
			
	}
fail:
	printk(KERN_ERR "vblank: Unable to map card.\n");
	return -ENOMEM;
}

static void vblankdev_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct video_card *card = dev_id;
	
	spin_lock(&card->lock);
	if(card->vops->test(card))
	{
		card->count++;
		wake_up_interruptible(&card->wait);
		kill_fasync(&card->async_queue, SIGIO, POLL_IN);
	}
	spin_unlock(&card->lock);
}

static int vblankdev_open(struct inode *inode, struct file *filp)
{
	struct video_card *card;
	card = kmalloc(sizeof(struct video_card), GFP_KERNEL);
	if(card == NULL)
		return -ENOBUFS;
	memset(card, 0, sizeof(struct video_card));
	init_waitqueue_head(&card->wait);
	spin_lock_init(&card->lock);
	filp->private_data = filp;
	return 0;
}

static int vblankdev_fasync (int fd, struct file *filp, int mode)
{
	struct video_card *card = filp->private_data;
	return fasync_helper(fd, filp, mode, &card->async_queue);
}

static int vblankdev_release(struct inode *inode, struct file *filp)
{
	struct video_card *card = filp->private_data;
	
	if(card->running)
	{
		card->vops->disable(card);
		free_irq(card->pci_dev->irq, card);
	}
	if(card->mmio)
		iounmap((void *)card->mmio);
	vblankdev_fasync(-1, filp, 0);
	kfree(card);
	return 0;
}

static unsigned int vblankdev_poll(struct file *filp, poll_table *poll_table)
{
	struct video_card *card = filp->private_data;
	
	poll_wait(filp, &card->wait, poll_table);
	if(card->count)
		return POLLIN;
	return 0;
}

static ssize_t vblankdev_read(struct file *filp, char *buff, size_t count, loff_t *offp)
{
	struct video_card *card = filp->private_data;
	unsigned long flags;
	int err = -EWOULDBLOCK;
	u32 num;
	size_t len;
	DECLARE_WAITQUEUE(wait, current);

	/*  Can't seek (pread) on this device  */
	if (offp != &filp->f_pos)
		return -ESPIPE;


	add_wait_queue(&card->wait, &wait);
	do
	{
		set_current_state(TASK_INTERRUPTIBLE);
		spin_lock_irqsave(&card->lock, flags);
		if(card->running)
		{
			num = card->count;
			card->count = 0;
			err = 0;
			break;
		}
		spin_unlock_irqrestore(&card->lock, flags);
		schedule();
	}
	while(!(filp->f_flags & O_NDELAY));
	
	remove_wait_queue(&card->wait, &wait);
	set_current_state(TASK_RUNNING);
	
	if(err)
		return err;
	
	len = min(count, (size_t)4);
	memcpy(buff, &num, 4);
	return len;
}

static int vblankdev_do_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct vblank_bind vbind;
	struct pci_dev *pdev;
	struct video_card *card = filp->private_data;

	switch(cmd)
	{	
		case VBLIOC_BIND:
			if(card->running)
				return -EBUSY;
				
			if(copy_from_user(&vbind, (void *)arg, sizeof(struct vblank_bind)))
				return -EFAULT;
		
			/* Domain isnt used yet */
			pdev = pci_find_slot(vbind.bus, vbind.devfn);
			if(pdev == NULL)
				return -ENODEV;
	
			if(pci_enable_device(pdev) < 0)
				return -EIO;
	
			if(pdev->irq == 0)
				return -EOPNOTSUPP;
	
			/* Ok try and bind it */
			card->pci_dev = pdev;
			if(vga_init_vsync(card) < 0)
				return -EOPNOTSUPP;
			card->running = 1;
			if(request_irq(pdev->irq, vblankdev_interrupt, SA_SHIRQ, "vblank", card) < 0)
			{
				card->running = 0;
				return -EBUSY;
			}
			card->vops->enable(card);
			return 0;
		default:
			return -ENOTTY;
	}
}

/*
 *	We need a lock here to avoid parallel binding of the same
 *	object.
 */

static DECLARE_MUTEX(ioctl_sem);

static int vblankdev_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int ret;
	
	down(&ioctl_sem);
	ret = vblankdev_do_ioctl(inode, filp, cmd, arg);
	up(&ioctl_sem);
	
	return ret;
}

static struct file_operations vblankdev_fops = {
	owner:		THIS_MODULE,
	llseek:		no_llseek,
	read:		vblankdev_read,
	ioctl:		vblankdev_ioctl,
	open:		vblankdev_open,
	release:	vblankdev_release,
	fasync:		vblankdev_fasync,
	poll:		vblankdev_poll
};

static struct miscdevice vblankdev = {
	VBLANK_MINOR,
	"vblank",
	&vblankdev_fops
};

static int vblankdev_init(void)
{
	if(!pci_present())
		return -ENODEV;
	if(misc_register(&vblankdev) < 0)
		return -EINVAL;
		
	return 0;
}

static void vblankdev_exit(void)
{
	misc_deregister(&vblankdev);
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Soeren Sandmann (sandmann@daimi.au.dk), very much based on work by Matan Ziv-Av (matan@svgalib.org), Alan Cox");
MODULE_PARM(try_vga, "i");

module_init(vblankdev_init);
module_exit(vblankdev_exit);
